/**************************************************************************

    This program is copyright (C) 2000 Sam Trenholme

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

**************************************************************************/

#include <Pilot.h>				
#include <Day.h>
#include <SelDay.h>
#include <DateTime.h>
#include <Preferences.h>

#include "CountRsc.h"
#ifdef __GNUC__
#include "Callbacks.h"
#endif		

// for definition of IsDigit
#include <CharAttr.h>
#include <SysEvtMgr.h>

// Some defines for displaying the countdown
#define COUNTDOWN_Y 40
#define COUNTDOWN_HEIGHT 10

// Global variables
// Variables that start with 'CountPref' are preferences stored between
// sessions.  'CountGlobal' variables are global variables not saved
// between sessions.

ULong CountPrefEventDay; // The day the event we are counting days since

int CountGlobalDateFormat; // The date format the system uses

static Err CountdownStartApplication(void)
{

	ULong ReadPref[1];	
	Word size = sizeof(ULong);
	SystemPreferencesType sysPrefs; 

	ReadPref[0] = 0;

	// Get the information from the Preferences, if needed
	PrefGetAppPreferences('Lord',0,ReadPref,&size,true);

	// Since ReadPref[0] will have value of 0 if the above fails, we
	// don't need to error check here
	CountPrefEventDay = ReadPref[0];

	// Now, we need to get the system-wide preferences, because
	// we use one field from them (the date format)
	PrefGetPreferences (&sysPrefs);

	// Get the date format the user prefers
	CountGlobalDateFormat = sysPrefs.dateFormat;

	// Open up the main form
        FrmGotoForm(CountdownForm);

	return 0;
}

static void CountdownStopApplication(void)
{

	ULong SavePref[1];
	Word size = sizeof(ULong);

	SavePref[0] = CountPrefEventDay;

	// Set the information in the Preferences
	PrefSetAppPreferences('Lord',0,1,SavePref,sizeof(ULong),true);

	// Next, close the forms
	FrmCloseAllForms(); 

}

// When they ask for help, give them a help screen
void CountdownDoHelp()
{
	FrmAlert(CountdownHelpAlert);

}

// Clear out a rectanguar box
// Input: x: left hand side of box
//        y: top of box
//        width:  width of box
//        height: height of box
void CountdownClearBox(int x, int y, int width, int height)
{
        RectanglePtr erase_box;

        // They had better keep RectangleType the structure RectanglrPtr
        // points to
        if((erase_box = MemPtrNew(sizeof(RectangleType))) == NULL)
                {
                // Is there a better way to handle memory starvation?
                FrmAlert(CountdownOutOfMemory);
                return;
                }

        // Initialize a rectangle the size of the text field we need to erase
        RctSetRectangle(erase_box,x,y,width,height);

        // And erase the text field     
        WinEraseRectangle(erase_box,0);

        // Now, free the memory we allocated above
        MemPtrFree(erase_box);

}

// This converts the value in CountPrefEventDay in to a string like
// 03/12/1987
char * CountdownDateStr(char *string)
{

	DateTimeType convert;

	// Obligatory fatal exception protection
	if(string == NULL)
		return NULL;

	// Get the month, day, and year
	TimSecondsToDateTime(CountPrefEventDay,&convert);

	// Overflow protection
	// I don't make assumptions about the data I get
	if(convert.month > 3000 || convert.month < 0) 
		convert.month = 0;	
	if(convert.day > 3000 || convert.month < 0) 
		convert.day = 0;	
	if(convert.year > 3000 || convert.month < 0) 
		convert.year = 0;	

	// There is a nice system routine for converting this
	// I prefer 03/12/1987 over 3/12/87 myself, so maybe I should
	// write my own form of DateToAscii
	DateToAscii(convert.month,convert.day,convert.year,
	            CountGlobalDateFormat,string);

	return string;
}

// CountdownDrawString: Draws a specified string that is centered at the
//                      user-specified height.  This takes up an entire
//	                line on the PalmOS display
// Input: string: a pointer to a pre-allocated string to draw
// y: the location (up-down) to draw the string
// Note that this code is hard wired to assume the display is 160 characters
// wide.
void CountdownDrawString(char *string, int y)
	{

	int width, strlen; // The width of the text, the string length of text
	int x; // Where to put the text

	// Determine the length of the string
	strlen = StrLen(string);
	
	// Some overflow protection
	if(strlen > 96)
		strlen = 96;

	// Clear the area where the string will be displayed
	CountdownClearBox(1,y,159,COUNTDOWN_HEIGHT);

	// Determine where to place the text so it lookes centered;
	width = FntCharsWidth(string,strlen);
	if(width > 158)
		width = 158;
	x = 80 - (width / 2);
	
	// And display the string in question
	WinDrawChars(string,strlen,x,y);

	// Exit the function
	return;

}

// When they ask how many days since the given event, either do nothing
// if the day has not been set, or tell them the number of days
// elapsed
void CountdownShowDays()
{
	ULong days,now;

	char string[32];
	char datestr[dateStringLength + 2];

	int sign = 0; // Sign of days; 0: positive; 1: negative

	// Return if day has not been set
	if(CountPrefEventDay == 0)
		return;

	// Determine if the event is before of after now, and set days
	// and sign accordingly
	now =  TimGetSeconds();
	// If the event was in the past
	if(now > CountPrefEventDay)
		days = TimGetSeconds() - CountPrefEventDay;
	// Otherwise, it will be in the future
	else
		{
		days = CountPrefEventDay - TimGetSeconds();
		days += 86400; // We want to round up, not round down when
			       // looking in to the future
		sign = 1;
		}

	// Since the value is currently in seconds, make it be actual days	
	days /= 86400; // Number of seconds in a day
	
	if(days >= 1000000 || days < 0)  // Overflow protection
		days = 0;

	// Convert the number to a string
	StrIToA(string,days);

	// add day/days
	if(days != 1)
		StrNCat(string," days ",45);
	else
		StrNCat(string," day ",45);

	// Add the "before/since event" string
	if(sign) 
		StrNCat(string,"before ",45);
	else
		StrNCat(string,"since ",45);

	// Add the date of the event to the displayed line
	CountdownDateStr(datestr);

	// Limit the buffer that can be copied, fatal exception protection
	datestr[dateStringLength] = 0;

	// Add the date 
	StrNCat(string,datestr,31);

	// Draw the string in question
	CountdownDrawString(string,COUNTDOWN_Y);

	// And exit from the function
	return;

}

// When they ask how many years, months, and days since the given event, 
// either do nothing if the day has not been set, or tell them the number 
// of years, months, and days elapsed
void CountdownShowYMD()
{
	ULong now,carrie;

	int years, months, days, c_month, c_year;

	char string[46];
	char datestr[dateStringLength + 2];
	char numconvert[8];

	int sign = 0; // Sign of days; 0: positive; 1: negative

	DateTimeType today,event;

	// Return if day has not been set
	if(CountPrefEventDay == 0)
		return;

	now = TimGetSeconds();

	// Get the month, day, and year for both events
	TimSecondsToDateTime(CountPrefEventDay,&event);
	TimSecondsToDateTime(now,&today);

	// Determine if the event is in the past or the future, and act 
        // accordingly
	// If the event was in the past
	if(now > CountPrefEventDay)
		{
		years = today.year - event.year;
		months = today.month - event.month;
		days = today.day - event.day;
		}
	// Otherwise, it will be in the future
	else
		{
		years = event.year - today.year;
		months = event.month - today.month;
		days = event.day - today.day;
		sign = 1;
		}

	// We have to carry over if years, months, or days is negative

	// Doing the carry over for days is the most tricky part
	// basically, we have to add the number of days we had last month
	// when carrying over here (or one month before event for future 
        // events)
	if(days < 0)
		{
		months--;

		if(!sign)
			{
			c_month = today.month;
			c_year  = today.year;
			}
		else
			{
			c_month = event.month;
			c_year  = event.year;
			}

		if(c_month > 1)
			c_month--;
		else	
			{
			c_month = 12;
			c_month--;
			}

		carrie = DaysInMonth(c_month,c_year);
		days += carrie;

		}
	if(months < 0)
		{
		months += 12;
		years--;
		}
		
	if(days >= 10000 || days < 0)  // Overflow protection
		days = 36;
	if(months >= 10000 || months < 0)  // Overflow protection
		months = 15;
	if(years >= 10000 || years < 0)  // Overflow protection
		years = 0;

	// Initialize the string;
	string[0] = '\0';

	// Convert the number of years to a string
	StrIToA(string,years); // we can use string since this is at the begin

	// add year/years
	if(years != 1)
		StrNCat(string," years ",45);
	else
		StrNCat(string," year ",45);

	// Convert the number of months to a string
	StrIToA(numconvert,months);
	
	// add that to the output string
	StrNCat(string,numconvert,45);

	// add month/months
	if(months != 1)
		StrNCat(string," months ",45);
	else
		StrNCat(string," month ",45);

	// Convert the number of months to a string
	StrIToA(numconvert,days);
	
	// add that to the output string
	StrNCat(string,numconvert,45);

	// add day/days
	if(days != 1)
		StrNCat(string," days ",45);
	else
		StrNCat(string," day ",45);

	// Draw the string in question
	CountdownDrawString(string,COUNTDOWN_Y + COUNTDOWN_HEIGHT + 3);

	// And exit from the function
	return;

}

// When they request to reset the day the even happened, ask them if they
// are sure they want to do this, then reset the day if they say yes
void CountdownSetDay()
{

	SWord Month, Day, Year;
	DateTimeType convert;

	// Some vaguely sane preferences just in case
	convert.month = 3;
	convert.day   = 12;
	convert.year  = 1987;

	// If the value is not set, default to today, otherwise have
	// the currently selected day be the day of the event
	if(CountPrefEventDay == 0)
		TimSecondsToDateTime(TimGetSeconds(),&convert);
	else
		TimSecondsToDateTime(CountPrefEventDay,&convert);

	Month = convert.month;
	Day   = convert.day;
	Year  = convert.year;	

	// Get the information via the SelectDay dialog
	SelectDay(selectDayByDay,&Month,&Day,&Year,"Day of event");

	// Convert the value obtined above to Palm-time()-style format,
	// and save the value in question to CountPrefEventDay
	convert.second = 0; // Note the time is midnight for the day.
	convert.hour = 0;   // We can get away with this because the Palm
	convert.minute = 0; // has no concept of time zones.
	convert.day = Day;
	convert.month = Month;
	convert.year = Year;
	convert.weekDay = 69; // Make this an invalid value

	// Now, convert, and save the value
	CountPrefEventDay = TimDateTimeToSeconds(&convert);
	
	// Just in case they mess around and choose January 1, 1904
	// This way we don't incorrectly tell them the day was not set
	// It is off for one second at midnight for this special case,
	// but that is a non-issue
	if(CountPrefEventDay == 0)
		CountPrefEventDay = 1;

	return;

}

// Callback to handle all the events on Countdown's single form
static Boolean CountdownHandleEvent(EventPtr event)
{
	Boolean		handled;

#ifdef __GNUC__
	CALLBACK_PROLOGUE
#endif
	handled = false;

	switch (event->eType)
		{

   		case ctlSelectEvent:  // Button was pressed and released.
			// Just in case we ever add any other buttons
			switch(event->data.ctlSelect.controlID) 
				{
				case CountdownSetDayButton:
					CountdownSetDay();
					CountdownShowDays();
					CountdownShowYMD();
					handled = true;
					break;
				}
			break;

		// This is run at startup time
  		case frmOpenEvent:	
			// Physically draw the form
			FrmDrawForm(FrmGetActiveForm());

			// Raz them to set the time if this is the first
			// time they have run the program
			if(CountPrefEventDay == 0)
				FrmAlert(CountdownNoDaySet);
			else
			// Show them the number of days since the "event"
				{
				CountdownShowDays();
				CountdownShowYMD();
				}

			handled = true;
			break;
		
		// Has the user selected something on the menu on top?	
		case menuEvent:		
			if (event->data.menu.itemID == CountdownMenuAbout)
   				FrmAlert(CountdownAboutAlert);
			else
				CountdownDoHelp();
			handled = true;
			break;

		}
#ifdef __GNUC__
	CALLBACK_EPILOGUE
#endif
	return(handled);
}


static Boolean CountdownHandleEventWrapper(EventPtr event)
{
	FormPtr	frm;
	Int		formId;
	Boolean	handled = false;

	if (event->eType == frmLoadEvent)
		{
		// Load the form resource specified in the event then activate the form.
		formId = event->data.frmLoad.formID;
		frm = FrmInitForm(formId);
		FrmSetActiveForm(frm);

		// Set the event handler for the form.  The handler of the currently 
		// active form is called by FrmDispatchEvent each time it receives an event.
		switch (formId)
			{
			case CountdownForm:
				FrmSetEventHandler(frm, CountdownHandleEvent);
				break;
			}
		handled = true;
		}
	
	return handled;
}

static void CountdownEventLoop(void)
{
	EventType	event;
	Word		error;
	
	do
		{
		EvtGetEvent(&event, evtWaitForever);
		if (! SysHandleEvent(&event))
			if (! MenuHandleEvent(0, &event, &error))
				if (! CountdownHandleEventWrapper(&event))
					FrmDispatchEvent(&event);
		}
	while (event.eType != appStopEvent);
}


DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
	Err err = 0;		
	
	if (cmd == sysAppLaunchCmdNormalLaunch)
		{
		if ((err = CountdownStartApplication()) == 0)
		{
			CountdownEventLoop();
			
			CountdownStopApplication();
			}
		}

	// Close all forms to stop memory leaks

	return err;
}

